import traceback
import requests
import shutil
import os
import logging
logging.getLogger('matplotlib').setLevel(logging.WARNING)

import redis
import numpy as np

from PIL import Image
from PIL import ImageFile
ImageFile.LOAD_TRUNCATED_IMAGES = True

import skimage
from activitystreams import parse as as_parser, Activity
from gnenv.environ import GNEnvironment
from flask import request
from abc import ABC
from abc import abstractmethod


class DownloadError(Exception):
    def __init__(self, status_code):
        self.status_code = status_code


class ConfigKeys:
    REDIS_HOST = 'redis_host'
    REDIS_PORT = 'redis_port'
    REDIS_DB = 'redis_db'
    LIB_PATH = 'lib_path'


class Detector(ABC):
    def __init__(self, env: GNEnvironment, logger, download_base: str, cf_key: str):
        self.logger = logger
        self.env = env
        self.download_base = download_base
        self.download_path = self.download_base + '/{}'
        self.cf_capacity = 100_000_000
        self.cf_fp_size = 2  # 0.01% error
        self.cf_key = cf_key
        self.use_filter = False
        self.redis_pool = None
        self.redis_instance = None

        self.lib_path = env.config.get(
            ConfigKeys.LIB_PATH, default='/var/lib/redis/libredis-cuckoofilter.so')

        self.create_download_directory()

        self.init_redis(env)
        self.init_cuckoo_filter(env)

    def init_redis(self, env: GNEnvironment):
        host = env.config.get(ConfigKeys.REDIS_HOST)
        port = env.config.get(ConfigKeys.REDIS_PORT)
        db = env.config.get(ConfigKeys.REDIS_DB)

        self.redis_pool = redis.ConnectionPool(host=host, port=port, db=db)
        self.redis_instance = None

    def init_cuckoo_filter(self, env:GNEnvironment):
        loaded_modules = self.redis.execute_command('module', 'list')
        already_loaded = False

        for loaded_module in loaded_modules:
            if str(loaded_module[1], 'utf-8') == 'cuckoofilter':
                already_loaded = True
                break

        if not already_loaded:
            try:
                self.redis.execute_command('module', 'load', self.lib_path)
            except Exception as e:
                self.logger.error(f'could not load cuckoofilter redis module: {str(e)}')
                self.logger.exception(e)
                return

        # the filter expects values such as '256M', '8G' etc., 'sizefor' calculates this based on fp size and capacity
        size = self.redis.execute_command('cf.sizefor', str(self.cf_capacity), str(self.cf_fp_size))
        size = str(size, 'utf-8')

        try:
            self.redis.execute_command('cf.init', self.cf_key, size, str(self.cf_fp_size))
            self.logger.info('filter does not exist, creating new')
        except redis.exceptions.ResponseError:
            self.logger.info('filter already exists, reusing')

        self.use_filter = True

    @property
    def redis(self):
        if self.redis_pool is None:
            return self.redis_instance
        return redis.Redis(connection_pool=self.redis_pool)

    def post(self):
        try:
            json_data = request.get_json()
        except Exception as e:
            error_msg = 'error: could not get json data from request: %s' % str(e)
            self.logger.error(error_msg)
            self.logger.exception(traceback.format_exc(e))
            return None

        try:
            data = self.do_post(json_data)
        except DownloadError as e:
            self.logger.error(f'could not download file, response code: {e.status_code}')
            return self.fail(json_data, 'could not download file', error_code=e.status_code)
        except Exception as e:
            self.logger.error('post failed: {}'.format(str(e)))
            self.logger.exception(traceback.format_exc())
            return self.fail(json_data, f'failed to do detection: {str(e)}', error_code=500)

        return self.ok(data)

    def do_post(self, json_data):
        activity = as_parser(json_data)
        dupe_hash, dupe_fingerprint = None, None
        is_dupe = False

        try:
            dupe_hash, dupe_fingerprint = self.generate_hash_and_fingerprint(activity)
        except Exception as e:
            self.logger.error(f'could not generate hash/fingerprint, ignoring dupe check: {str(e)}')
            self.logger.exception(e)

        try:
            is_dupe = self.is_duplicate(dupe_hash, dupe_fingerprint)
        except Exception as e:
            self.logger.error(f'could not check if duplicate: {str(e)}')
            self.logger.exception(e)

        if is_dupe:
            self.logger.info(f'ignoring duplicate request with hash {dupe_hash} and fingerprint {dupe_fingerprint}')
            self.logger.info(json_data)
            return self.fail(json_data, msg='duplicate request', error_code=422)

        image_url = activity.object.url
        image_path = self.download(image_url)

        detection_results = self.detect(image_path)
        response = self.create_response(detection_results)

        self.logger.info(f'sending response: {response}')

        try:
            self.add_to_filter(response, dupe_hash, dupe_fingerprint)
        except FileNotFoundError as e:
            self.logger.warning(f'could not add {dupe_hash}/{dupe_fingerprint} to cuckoo filter: {str(e)}')

        try:
            os.remove(image_path)
        except FileNotFoundError as e:
            self.logger.warning(f'could not remove file {image_path}, ignoring: {str(e)}')

        return response

    def is_duplicate(self, img_hash: int, img_fingerprint: int) -> bool:
        if not self.use_filter:
            return False

        exists = self.redis.execute_command('cf.check', self.cf_key, str(img_hash), str(img_fingerprint))
        if exists is None:
            return False

        if int(str(exists, 'utf-8')) == 1:
            return True
        return False

    def add_to_filter(self, response: dict, dupe_hash: int, dupe_fingerprint: int):
        if response is None or dupe_hash is None or dupe_fingerprint is None:
            return
        if 'content' not in response or response['content'] != '200':
            return

        self.redis.execute_command('cf.add', self.cf_key, str(dupe_hash), str(dupe_fingerprint))

    @abstractmethod
    def generate_hash_and_fingerprint(self, activity: Activity) -> (int, int):
        pass

    @abstractmethod
    def detect(self, image_path: str):
        pass

    @abstractmethod
    def create_response(self, detection_results):
        pass

    @abstractmethod
    def fail(self, json_data, msg, error_code=None):
        pass

    def ok(self, data):
        return data

    def download(self, image_url):
        out_file_path = self.download_path.format(image_url.split('/')[-1])
        self.logger.info(f'saving {image_url} to {out_file_path}')

        response = requests.get(image_url, stream=True)

        if response.status_code != 200:
            raise DownloadError(response.status_code)

        self.create_download_directory()
        with open(out_file_path, 'wb') as out_file:
            shutil.copyfileobj(response.raw, out_file)

        return out_file_path

    def create_download_directory(self):
        if not os.path.exists(self.download_base):
            os.mkdir(self.download_base, mode=0o775)

    def load_image(self, image_file):
        """
        Sometimes skimage will hang when loading an image in gpu-02, but not on gpu-01.
        Internally skimage uses PIL to open images, but somewhere it gets stuck. Using
        PIL directly seems to work around the issues, so try that first. The only downside is
        that we have to do the conversion to np arrays manually and for some image formats
        this might not work, cause some corner case is not covered, so try to use skimage
        when using PIL fails.
        """
        try:
            image = Image.open(image_file)
            return self.pillow2array(image)
        except Exception as e:
            self.logger.error(f'[PIL] could not open image {image_file}, trying with skimage: {str(e)}')

        try:
            return skimage.io.imread(image_file)
        except Exception as e:
            self.logger.error(f'[skimage] could not open image {image_file}, giving up: {str(e)}')
            raise e

    def pillow2array(self, img):
        arr = np.array(img.getdata(), np.uint8)
        return arr.reshape(
            img.size[1],
            img.size[0],
            len(img.getbands())  # might be greyscale (1 channel), RGB (3 channels) or transparent (4 channels)
        )
