# detectorlib

Provides a base class for classifier models that are getting request from Logistik. Enables single-file serving of 
models. Provides the following functionality:

* request parsing (ActivityStreams),
* image downloading,
* cleanup after request has been processed,
* image loading and resizing,
* handles control flow of requests (retry, dropping, etc.),
* duplicate request checking,
* model filtering (not all models are always needed),
* exposes config values to Logistik,
* exposes log file to Logistik UI. 

## Installation

Either install using pip directly:

```bash
pip install git+http://gitlab.thenetcircle.lab/tnc-service-team/detectorlib.git
```

Or add to your `requirements.txt` file:

```bash
# cat requirements.txt
activitystreams==0.2.1
git+http://gitlab.thenetcircle.lab/tnc-service-team/detectorlib.git
PyYAML==5.3
# pip install -r requirements.txt
```

## Usage:

Extending the `Detector` base class can be done like the following:

```python
# other imports

from keras.models import load_model as keras_load_model

from gnenv import create_env      # handles config file and secrets
from detectorlib import Detector  # base class for classifiers


class ThingsDetector(Detector):
    def __init__(self):
        super().__init__(
            env=create_env(),
            logger=logging.getLogger(__name__),
            # detectorlib handles image downloading
            download_base='/tmp/things-downloads',
            # cockoo filter key to use, detectorlib will drop duplicate requests
            cf_key=f'cf.things.popp.prod.v1',
            # detectorlib will discard any requests not specifying this channel
            channel='things'
        )
        self.model = keras_load_model(self.env.config.get(ConfigKeys.MODEL_PATH))

    # overriding abstract method
    def detect(self, image_path):
        image = self.load_image(image_path, dims=384)
        return self.model.predict(image)[0][0]

    # overriding abstract method
    def create_response(self, activity: Activity, y_hat):
        return ActivityBuilder.enrich(activity, y_hat=y_hat)

    # overriding abstract method
    def fail(self, json_data, msg, error_code=None):
        return ActivityBuilder.enrich(as_parser(json_data), message=msg, e_code=error_code), error_code


app = Flask(__name__)
detector = ThingsDetector()


# (Optional) logistik will query /info to get config values such as return_to topic, kafka group_id etc
@app.route('/info/<environment>', methods=['GET'])
def info(environment: str):
    return jsonify(detector.info(environment))


# (Optional) logistik frontend can request to see the tail of the log file
@app.route('/log', methods=['GET'])
def tail_log_file():
    response, status_code = detector.logs()
    return jsonify(response), status_code


# (Required) the api to call to classify an image
@app.route(env.config.get(ConfigKeys.API_PATH), methods=['POST'])
def detect():
    response, status_code = detector.post()
    return jsonify(response), status_code
```