import traceback
from functools import wraps

import requests
import shutil
import os
import subprocess
import logging
import ast
import easytracer

logging.getLogger("matplotlib").setLevel(logging.WARNING)

import redis
import numpy as np

from PIL import Image
from PIL import ImageFile

ImageFile.LOAD_TRUNCATED_IMAGES = True

import skimage
from activitystreams import parse as as_parser, Activity
from gnenv.environ import GNEnvironment, find_config, load_secrets_file
from flask import request
from abc import ABC
from abc import abstractmethod


class DownloadError(Exception):
    def __init__(self, status_code):
        self.status_code = status_code


class ConfigKeys:
    REDIS_HOST = "redis_host"
    REDIS_PORT = "redis_port"
    REDIS_DB = "redis_db"
    LIB_PATH = "lib_path"
    MODEL_PATH = "model_path"
    CF_CAPACITY = "cf_capacity"
    LOG_FILE = "log_file"
    API_PATH = "path"
    ANCHORS_PATH = "anchors_path"
    CLASSES_PATH = "classes_path"
    WIDTH = "width"
    HEIGHT = "height"
    SCORE = "score"
    IOU = "iou"
    GROUP_ID = "group_id"
    DETECTION_TYPE = "detection_type"


provider_to_int = {
    "popcorn": "2",
    "poppen": "2",
    "fetish": "3",
    "gays": "4",
    "fetischde": "5",
    "fet": "6",
    "koko": "7",
    "gaudi": "8",
}


def trace(tracer, ctx, name):
    def factory(view_func):
        @wraps(view_func)
        def decorator(*args, **kwargs):
            # implementing service has not enabled tracing
            if ctx is None or tracer is None:
                return view_func(*args, **kwargs)

            with tracer.start_span(name=name, child_of=ctx) as _:
                return view_func(*args, **kwargs)

        return decorator
    return factory


class Detector(ABC):
    def __init__(
        self,
        env: GNEnvironment,
        logger,
        download_base: str,
        cf_key: str,
        logfile: str = None,
        log_tail_lines: int = 5000,
        channel: str = None,
        use_filter: bool = False
    ):
        self.logger = logger
        self.env = env
        self.download_base = download_base
        self.download_path = self.download_base + "/{}"
        self.cf_capacity = env.config.get(ConfigKeys.CF_CAPACITY)
        self.logfile = logfile
        self.log_tail_lines = log_tail_lines
        self.cf_fp_size = 2  # 0.01% error
        self.cf_key = cf_key
        self.channel = channel
        self.use_filter = use_filter
        self.redis_pool = None
        self.redis_instance = None
        self.tracer = None
        self.lib_path = env.config.get(
            ConfigKeys.LIB_PATH, default="/var/lib/redis/libredis-cuckoofilter.so"
        )

        self.create_download_directory()

        self.init_redis()
        self.init_cuckoo_filter()
        self.init_logfile()

    def post(self):
        try:
            json_data = request.get_json()
        except Exception as e:
            error_msg = "error: could not get json data from request: %s" % str(e)
            self.logger.error(error_msg)
            self.logger.exception(traceback.format_exc(e))
            return None, 400

        try:
            data, code = self.do_post(json_data)
        except DownloadError as e:
            self.logger.error(
                f"could not download file, response code: {e.status_code}"
            )
            fail_activity = self.fail(
                json_data, "could not download file", error_code=e.status_code
            )
            return fail_activity, e.status_code
        except Exception as e:
            self.logger.error("post failed: {}".format(str(e)))
            self.logger.exception(traceback.format_exc())
            fail_activity = self.fail(
                json_data, f"failed to do detection: {str(e)}", error_code=500
            )
            return fail_activity, 500

        return data, code

    def get_tracing_span_context(self):
        try:
            return easytracer.extract(request.headers)
        except Exception as e:
            logging.error("could not extract tracing info from headers: {}".format(str(e)))

    def request_is_for_this_channel(self, activity: Activity):
        if self.channel is None or len(self.channel.strip()) == 0:
            return True

        if not hasattr(activity, "target"):
            return True

        target = activity.target
        if not hasattr(target, "content") or not hasattr(target, "object_type"):
            return True

        if target.object_type != "channel":
            return True
        if target.content is None or len(target.content) == 0:
            return True

        try:
            channels = ast.literal_eval(target.content)
        except Exception as e:
            self.logger.error(f"count not check target.content: {str(e)}")
            self.logger.error(f"target.content was {target.content}")
            self.logger.exception(e)
            return True

        # evaluating "[]" will have length 0
        if len(channels) == 0 or not type(channels) == list:
            return True

        # evaluating "[\"\"]" will have length 1, and first element length 0
        if len(channels) == 1 and len(channels[0]) == 0:
            return True

        is_for_this_channel = self.channel in channels

        if not is_for_this_channel:
            self.logger.info(
                f"request is for channels {channels} but this model is for {self.channel}, ignoring"
            )

        return is_for_this_channel

    def do_post(self, json_data):
        activity = as_parser(json_data)
        dupe_hash, dupe_fingerprint = None, None
        is_dupe = False

        if not self.request_is_for_this_channel(activity):
            return dict(), 204

        try:
            dupe_hash, dupe_fingerprint = self.generate_hash_and_fingerprint(activity)
        except Exception as e:
            self.logger.error(
                f"could not generate hash/fingerprint, ignoring dupe check: {str(e)}"
            )
            self.logger.exception(e)

        # some services might want to ignore checking for duplicates, e.g. fakecheck-service
        if self.should_skip_duplication_check(activity):
            is_dupe = False
        else:
            try:
                is_dupe = self.is_duplicate(dupe_hash, dupe_fingerprint)
            except Exception as e:
                self.logger.error(f"could not check if duplicate: {str(e)}")
                self.logger.exception(e)

        if is_dupe:
            self.logger.info(
                f"ignoring duplicate request with hash {dupe_hash} and fingerprint {dupe_fingerprint}"
            )
            self.logger.info(json_data)
            return self.fail(json_data, msg="duplicate request", error_code=422), 422

        span_ctx = self.get_tracing_span_context()

        image_url = activity.object.url
        image_path = self.download(image_url, span_ctx=span_ctx)

        detection_results = self.trace_detection(image_path, activity, span_ctx=span_ctx)
        response = self.create_response(activity, detection_results)

        if not self.should_skip_duplication_check(activity):
            self.add_to_filter(response, dupe_hash, dupe_fingerprint)

        try:
            os.remove(image_path)
        except FileNotFoundError as e:
            self.logger.warning(
                f"could not remove file {image_path}, ignoring: {str(e)}"
            )

        self.logger.info(f"sending response: {response}")
        return self.ok(response), 200

    def trace_detection(self, image_path, activity, span_ctx):
        @trace(self.tracer, span_ctx, "detect")
        def _detect():
            return self.detect_with_activity(image_path, activity)

        return _detect()

    def is_duplicate(self, img_hash: int, img_fingerprint: int) -> bool:
        if not self.use_filter:
            return False

        exists = self.redis.execute_command(
            "cf.check", self.cf_key, str(img_hash), str(img_fingerprint)
        )
        if exists is None:
            return False

        if int(str(exists, "utf-8")) == 1:
            return True
        return False

    def add_to_filter(self, response: dict, dupe_hash: int, dupe_fingerprint: int):
        if response is None or dupe_hash is None or dupe_fingerprint is None:
            return
        if "content" not in response or response["content"] != "200":
            return

        try:
            self.redis.execute_command(
                "cf.add", self.cf_key, str(dupe_hash), str(dupe_fingerprint)
            )
        except Exception as e:
            self.logger.error(f"could not add to filter, skipping: {str(e)}")

    def should_skip_duplication_check(self, activity: Activity) -> bool:
        if not hasattr(activity, "target"):
            return False
        if not hasattr(activity.target, "attachments"):
            return False
        if activity.target.attachments is None or len(activity.target.attachments) == 0:
            return False

        try:
            for attachment in activity.target.attachments:
                if attachment.object_type != "skip_dupe_check":
                    continue

                if int(float(attachment.content)) == 1:
                    return True
                else:
                    return False
        except Exception as e:
            self.logger.warning(
                f"could not check 'skip_dupe_check' on target: {str(e)}"
            )
            self.logger.exception(traceback.format_exc())

        return False

    def generate_hash_and_fingerprint(self, activity: Activity) -> (str, str):
        image_id = activity.object.id
        user_id = activity.actor.id
        provider_id = "1"

        if hasattr(activity, "provider") and hasattr(activity.provider, "id"):
            provider_id = activity.provider.id

        if provider_id in provider_to_int:
            provider_id = provider_to_int[provider_id]

        cf_hash = provider_id + image_id
        cf_fingerprint = user_id[-2:]

        return cf_hash, cf_fingerprint

    def ok(self, data):
        return data

    def download(self, image_url, span_ctx=None):
        @trace(self.tracer, span_ctx, "download")
        def _download():
            out_file_path = self.download_path.format(image_url.split("/")[-1])
            self.logger.info(f"saving {image_url} to {out_file_path}")

            response = requests.get(image_url, stream=True)

            if response.status_code != 200:
                raise DownloadError(response.status_code)

            self.create_download_directory()
            with open(out_file_path, "wb") as out_file:
                shutil.copyfileobj(response.raw, out_file)

            return out_file_path

        # need to wrap it so the decorator gets the span context
        return _download()

    def create_download_directory(self):
        if not os.path.exists(self.download_base):
            os.mkdir(self.download_base, mode=0o775)

    def load_image(self, image_file, dims=None, span_ctx=None):
        """
        Sometimes skimage will hang when loading an image in gpu-02, but not on gpu-01.
        Internally skimage uses PIL to open images, but somewhere it gets stuck. Using
        PIL directly seems to work around the issues, so try that first. The only downside is
        that we have to do the conversion to np arrays manually and for some image formats
        this might not work, cause some corner case is not covered, so try to use skimage
        when using PIL fails.
        """
        @trace(self.tracer, span_ctx, "load_image")
        def _load_image():
            try:
                image = Image.open(image_file)
                image = self.pillow2array(image)
            except Exception as e:
                self.logger.error(
                    f"[PIL] could not open image {image_file}, trying with skimage: {str(e)}"
                )
                try:
                    image = skimage.io.imread(image_file)
                except Exception as e2:
                    self.logger.error(
                        f"[skimage] could not open image {image_file}, giving up: {str(e2)}"
                    )
                    raise e2

            if dims is not None:
                try:
                    image = skimage.transform.resize(
                        image,
                        output_shape=(dims, dims, 3),
                        mode="reflect",
                        anti_aliasing=True,
                    )
                    image = np.expand_dims(image, axis=0)
                except Exception as e:
                    self.logger.error(
                        f"[skimage] could not transform image {image_file}: {str(e)}"
                    )
                    raise e

            return image

        # need to wrap it so the decorator gets the span context
        return _load_image()

    def pillow2array(self, img):
        arr = np.array(img.getdata(), np.uint8)
        return arr.reshape(
            img.size[1],
            img.size[0],
            len(img.getbands()),  # might be greyscale (1 channel), RGB (3 channels) or transparent (4 channels)
        )

    def info(self, environment: str = None):
        fields = [
            "return_to",
            "event",
            "method",
            "retries",
            "timeout",
            "group_id",
            "path",
            "failed_topic",
        ]
        response = dict()

        if environment is None:
            config_dict = self.env.config.params
        else:
            config_dict, config_path = find_config()
            config_dict = load_secrets_file(config_dict, env_name=environment)

        for field in fields:
            if field in config_dict.keys():
                response[field] = config_dict.get(field)

        return response

    def logs(self):
        if self.logfile is None:
            return list(), 200

        f = subprocess.Popen(
            ["tail", "-n", str(self.log_tail_lines), self.logfile],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        lines = list()
        while True:
            line = f.stdout.readline()
            if not line:
                break
            lines.append(str(line, "utf-8", "ignore"))

        return lines, 200

    def detect_with_activity(self, image_path: str, activity: Activity):
        """
        some models can overwrite if need be
        """
        return self.detect(image_path)

    @abstractmethod
    def detect(self, image_path: str):
        pass

    @abstractmethod
    def create_response(self, activity: Activity, detection_results):
        pass

    @abstractmethod
    def fail(self, json_data, msg, error_code=None) -> dict:
        """
        implementor decides fail format

        :param json_data:
        :param msg:
        :param error_code:
        :return: dict
        """

    def init_tracer(self, service_name):
        try:
            from easytracer import Config
        except ImportError:
            self.logger.warning("[detectorlib] easytracer is not installed, not enabling tracing")
            return

        config = Config(
            config={
                'sampler': {
                    'type': 'const',
                    'param': 1,
                },
                'logging': True
            },
            service_name=service_name
        )

        self.tracer = config.init_tracer()

    def init_logfile(self):
        if self.logfile is not None:
            return

        logfile = self.env.config.get(ConfigKeys.LOG_FILE)

        if os.path.exists("/".join(logfile.split("/")[:-1])):
            self.logfile = logfile
            self.logger.info(f"using logfile: {self.logfile}")
        else:
            self.logger.error(f"log file directory does not exist: {logfile}")

        default_log_tail_lines = 5000
        try:
            self.log_tail_lines = int(self.log_tail_lines)
            if self.log_tail_lines < 0 or self.log_tail_lines > 100000:
                self.logger.warning(
                    f"log_tail_lines is outsize the bounds of [0, 100000], using default of {default_log_tail_lines}"
                )
                self.log_tail_lines = default_log_tail_lines
        except ValueError:
            self.logger.warning(
                f"log_tail_lines is not a valid integer, using default of {default_log_tail_lines}"
            )
            self.log_tail_lines = default_log_tail_lines

    def init_redis(self):
        host = self.env.config.get(ConfigKeys.REDIS_HOST)
        port = self.env.config.get(ConfigKeys.REDIS_PORT)
        db = self.env.config.get(ConfigKeys.REDIS_DB)

        self.redis_pool = redis.ConnectionPool(host=host, port=port, db=db)
        self.redis_instance = None

    def init_cuckoo_filter(self):
        if not self.use_filter:
            return

        loaded_modules = self.redis.execute_command("module", "list")
        already_loaded = False

        for loaded_module in loaded_modules:
            if str(loaded_module[1], "utf-8") == "cuckoofilter":
                already_loaded = True
                break

        if not already_loaded:
            try:
                self.redis.execute_command("module", "load", self.lib_path)
            except Exception as e:
                self.logger.error(f"could not load cuckoofilter redis module: {str(e)}")
                self.logger.exception(e)
                return

        # the filter expects values such as '256M', '8G' etc., 'sizefor' calculates this based on fp size and capacity
        size = self.redis.execute_command(
            "cf.sizefor", str(self.cf_capacity), str(self.cf_fp_size)
        )
        size = str(size, "utf-8")

        try:
            self.redis.execute_command(
                "cf.init", self.cf_key, size, str(self.cf_fp_size)
            )
            self.logger.info("filter does not exist, creating new")
        except redis.exceptions.ResponseError:
            self.logger.info("filter already exists, reusing")

        self.use_filter = True

    @property
    def redis(self):
        if self.redis_pool is None:
            return self.redis_instance
        return redis.Redis(connection_pool=self.redis_pool)
