import logging
import time
from abc import ABC
from typing import List, Callable

from kafka import KafkaConsumer

ONE_MINUTE_MS = 60_000


class KafkaReader(ABC):
    def __init__(
        self,
        event_handler_method: Callable,
        bootstrap_servers: List,
        topic_name: str,
        group_id: str,
    ):
        self.logger = logging.getLogger(__name__)
        self.bootstrap_servers = bootstrap_servers
        self.topic_name = topic_name
        self.group_id = group_id
        self.event_handler_method: Callable = event_handler_method

        self.logger.info(f"bootstrapping from kafka servers: {self.bootstrap_servers}")
        self.logger.info(f"consuming from topic {self.topic_name}")

        # connections_max_idle_ms (540) must be larger than
        # request_timeout_ms (305000) which must be larger than
        # fetch_max_wait_ms (500).
        self.consumer = KafkaConsumer(
            self.topic_name,
            group_id=self.group_id,
            bootstrap_servers=self.bootstrap_servers,
            enable_auto_commit=True,
            auto_offset_reset="earliest",
            connections_max_idle_ms=9 * ONE_MINUTE_MS,  # default: 9min
            max_poll_interval_ms=10 * ONE_MINUTE_MS,  # default: 5min
            session_timeout_ms=ONE_MINUTE_MS,  # default: 10s
            max_poll_records=10,  # default: 500
        )

    def start_consuming(self, sleep_time=2, exit_on_failure=False):
        if sleep_time > 0:
            self.logger.info(f"sleeping for {sleep_time} second before consuming")
            time.sleep(sleep_time)

        while True:
            try:
                self.try_to_read()
            except InterruptedError:
                self.logger.info("got interrupted, shutting down...")
                break
            except Exception as e:
                self.logger.error(f"failed to handle message: {str(e)}")
                self.logger.exception(e)

                if exit_on_failure:
                    break

                time.sleep(1)

    def try_to_read(self):
        for message in self.consumer:
            try:
                # call implementing class' handler method
                self.event_handler_method(message)
            except InterruptedError:
                self.logger.warning(
                    f"got interrupt, dropping message: {str(message.value)}"
                )
                raise
